/* 
 Copyright 2005-2007 Foxmarks Inc.
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
        http://www.apache.org/licenses/LICENSE-2.0
        
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
 foxmarks-merge.js: implements the "initial merge" algorithm, where a
 foreign bookmarkset is merged into a base set.
 
 */
 
// merge two bookmark sets

function Merge(base, foreign)
{
    LogWrite("Starting merge");
        
    var dirty = false;
    var roots = [base.root, foreign.root];
    var toolbars = [base.toolbar, foreign.toolbar];
    var mergeToolbars = false;
    var resources = [roots];

    if (toolbars[0] && toolbars[1]) {
        LogWrite("Toolbars cool: " + base.ResourceName(toolbars[0]) + ", " + 
            foreign.ResourceName(toolbars[1]));
        resources.push(toolbars);
        mergeToolbars = true;
    }

    var folders = [];
        
    while (resources.length > 0) {
        dirty = false;
        folders = resources.shift();
        var basefolder = folders[0];
        var foreignfolder = folders[1];
                
        var basechildren = base.EnumerateChildren(basefolder);
        var foreignchildren = foreign.EnumerateChildren(foreignfolder);
                
        LogWrite("Processing " + foreign.ResourceName(foreignfolder));
                
        while (foreignchildren.length > 0) {
            item = foreignchildren.shift();
            if (mergeToolbars && item == foreign.toolbar) {
                LogWrite("Skipping toolbar " + foreign.ResourceName(item));
                continue;
            }

            match = MatchItem(base, foreign, basechildren, item);
                        
            if (match < 0) {                // no match
                // bug: sometimes folders (if empty?) aren't decorated as sequences
                if (!base.IsContainer(basefolder))
                    base.rdfcu.MakeSeq(base.ds, basefolder);
                                        
                base.container.Init(base.ds, basefolder);

                // careful: if the resource already exists, we'll
                // cause problems if we try to copy it again.
                                
                if (base.ResourceExists(item)) {
                    LogWrite("WARNING: " + foreign.ResourceName(item) + "already exists.");
                                        
                } else {
                                
                    if (!dirty) {
                        dirty = true;
                        base.MarkDirty(basefolder);
                    }
                                        
                    LogWrite("Didn't find match; appending " + foreign.ResourceName(item));
                    var id = base.CopyResourceAnonymous(foreign, item, mergeToolbars);
                    base.container.AppendElement(id);
                }
                                
            } else {                                                        // match
                LogWrite("Matched " + foreign.ResourceName(item) + " = " + 
                    base.ResourceName(basechildren[match]));
                if (foreign.IsContainer(item)) {
                    var matchedFolders = new Array(2);
                    matchedFolders[0] = basechildren[match];
                    matchedFolders[1] = item;
                    resources.push(matchedFolders);
                }
                                
                FixupMatchedItem(base, foreign, basechildren[match], item);
                // remove matched item from consideration for further matching
                basechildren.splice(match, 1);  
            }
        }       
    }
    LogWrite("Finished merge");
}

// match foreign: item against base:basechildren, returning index of match, -1 if no match
function MatchItem(base, foreign, basechildren, item)
{
    var rv = -1;
    const propType = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";
    const propName = "http://home.netscape.com/NC-rdf#Name";
    const propDesc = "http://home.netscape.com/NC-rdf#Description";
    const propURL = "http://home.netscape.com/NC-rdf#URL";
    const propKeyword = "http://home.netscape.com/NC-rdf#ShortcutURL";
    const propWebPanel = "http://home.netscape.com/NC-rdf#WebPanel";
    
    // first, I suppose we should check to see whether there's an item with the
    // same resource id. If there is, it automatically matches.
    rv = basechildren.indexOf(item);
    if (rv >= 0)
        return rv;
                
    var itemType = foreign.Prop(item, propType);
        
    if (itemType instanceof Components.interfaces.nsIRDFResource) {
        switch (itemType.Value) {
        case "http://home.netscape.com/NC-rdf#Folder":
            return MatchItemSpec(base, foreign, basechildren, item, [propType, propName, propDesc]);
        case "http://home.netscape.com/NC-rdf#Bookmark":
            return MatchItemSpec(base, foreign, basechildren, item, [propType, propURL]);
        case "http://home.netscape.com/NC-rdf#BookmarkSeparator":
            return MatchItemSpec(base, foreign, basechildren, item, [propType, propName]);
        default:
            return -1;
            // XXX What about Livemarks?
        }
    }
        
    return -1;
}

// tries to match a single item against one of a list of baseChildren,
// requiring that each of the provided predicates matches
// return the index into baseChildren of the first match, or -1 if no match found
function MatchItemSpec(base, foreign, baseChildren, item, predicates)
{
    var matched;
    var baseChild;
    var i, j;
        
    for (i = 0; i < baseChildren.length; ++i) {
        baseChild = baseChildren[i];
        matched = true;
        var predicate;
        for (j = 0; j < predicates.length; ++j) {
            predicate = predicates[j];
            if (base.Prop(baseChild, predicate) != foreign.Prop(item, predicate)) {
                matched = false;
                break;
            }
        }
        if (matched) {
            return i;
        }
    }
        
    return -1;
}

// If some of baseChild's attributes don't match foreignChild's,
// document the differences baseChild's description field.
function FixupMatchedItem(base, foreign, baseChild, foreignChild)
{
    // XXX Implement me!
}
